import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.UnknownHostException;
import java.net.InetAddress;
import java.util.Hashtable;
import java.util.logging.Logger;
import java.util.logging.Level;

public class KeyStorage {
	public static final int SET		= 0x01;
	public static final int GET		= 0x02;
	public static final int REMOVE	= 0x03;

	private static Logger logger = Logger.getLogger(KeyStorage.class.getName());
	private static final byte[] magic = { (byte) 0xba, (byte) 0xca };
	private static final int COMMAND_OFFSET = 2;
	private static final int HEADER_LENGTH = 3;
	private int channelID, myID;
	private DataInputStream din;
	private DataOutputStream dout;
	private Hashtable storage;

	private void send(ReflectorPacket rp) throws IOException {
		synchronized (dout) {
			dout.writeShort(rp.getBuffer().length);
			dout.write(rp.getBuffer());
			dout.flush();
		}
	}

	private void get(ReflectorPacket rp) throws IOException {
		byte[] buf = rp.getData();
		ByteArrayOutputStream bout = new ByteArrayOutputStream();
		for (int i = HEADER_LENGTH; i < buf.length && buf[i] != 0x00; i++) {
			bout.write(buf[i]);
		}
		String key = new String(bout.toByteArray(), "utf8");
		ReflectorPacket sendrp = new ReflectorPacket();
		sendrp.setCommand(Reflector.SEND);
		sendrp.setChannel(rp.getChannel());
		sendrp.setId(myID);
		sendrp.setTargets(new byte[] { (byte) rp.getId() });
		sendrp.setData((byte[]) storage.get(key));
		send(sendrp);
	}

	private void set(ReflectorPacket rp) throws IOException
			, UnsupportedEncodingException {
		byte[] buf = rp.getData();
		ByteArrayOutputStream bout = new ByteArrayOutputStream();
		int i = HEADER_LENGTH;
		for ( ; i < buf.length; i++) {
			if (buf[i] == 0x00) {
				i++;
				break;
			}
			bout.write(buf[i]);
		}
		String key = new String(bout.toByteArray(), "utf8");
		byte[] value = new byte[buf.length - i];
		System.arraycopy(buf, i, value, 0, value.length);
		storage.put(key, value);
	}

	private void remove(ReflectorPacket rp) throws IOException
			, UnsupportedEncodingException {
		byte[] buf = rp.getData();
		ByteArrayOutputStream bout = new ByteArrayOutputStream();
		for (int i = HEADER_LENGTH; i < buf.length && buf[i] != 0x00; i++) {
			bout.write(buf[i]);
		}
		String key = new String(bout.toByteArray(), "utf8");
		storage.remove(key);
	}

	private void process(ReflectorPacket rp) throws IOException
			, UnsupportedEncodingException {
		if (rp.getCommand() == Reflector.ENTRY) {
			myID = rp.getId();
			return;
		} else if (rp.getCommand() != Reflector.SEND) {
			logger.info("not SEND packet.");
			return;
		}
		byte[] buf = rp.getData();
		if (buf == null || buf.length == 0) {
			if (logger.isLoggable(Level.FINE))
				logger.fine("RECV=" + rp);
			return;
		}
		if (buf[0] != magic[0] || buf[1] != magic[1]) {
			logger.info("BAD magic.");
			return;
		}
		switch (buf[COMMAND_OFFSET]) {
		case GET:
			get(rp);
			break;
		case SET:
			set(rp);
			break;
		case REMOVE:
			remove(rp);
			break;
		default:
			break;
		}
	}

	private class KeepAlive extends Thread {
		private static final long INTERVAL = 25000L; // 25 sec.
		public void run() {
			try {
				while(true) {
					sleep(INTERVAL);
					ReflectorPacket rp = new ReflectorPacket();
					rp.setCommand(Reflector.SEND);
					rp.setChannel(channelID);
					rp.setId(myID);
					rp.setTargets(new byte[] { (byte) myID });
					send(rp);
				}
			} catch (Exception ex) {
				logger.log(Level.SEVERE, "KeepAlive thread died.", ex);
			}
		}
	}

	private class Receiver extends Thread {
		public void run() {
			try {
				while (true) {
					int length = din.readUnsignedShort();
					byte[] buf = new byte[length];
					din.readFully(buf);
					ReflectorPacket rp = new ReflectorPacket(buf, length);
					process(rp);
				}
			} catch (Exception ex) {
				logger.log(Level.SEVERE, "Receiver thread died.", ex);
			}
		}
	}

	public KeyStorage(int port, int channelid) throws UnknownHostException
			, IOException {
		storage = new Hashtable();
		channelID = channelid;
		Socket sock = new Socket(InetAddress.getLocalHost(), port);
		din = new DataInputStream(sock.getInputStream());
		dout = new DataOutputStream(sock.getOutputStream());
		new Receiver().start();
		// ENTRY
		ReflectorPacket rp = new ReflectorPacket();
		rp.setCommand(Reflector.ENTRY);
		rp.setChannel(channelID);
		send(rp);
		new KeepAlive().start();
	}
}
