import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.logging.Logger;
import java.util.logging.Level;

public class Reflector {
	/* commands */
	public static final int ENTRY		= 0x10;
	public static final int SEND		= 0x11;
	public static final int LIST		= 0x12;
	public static final int ADDRESS		= 0x13;
	public static final int EXIT		= 0x20;

	private static final int BUF_SIZE = 4096;
	private static final long CLEANING_INTERVAL = 5000L;	// 5 sec.
	private static final long CLIENT_TIMEOUT = 30000L;		// 30 sec.

	/* members */
	private static Logger logger = Logger.getLogger(Reflector.class.getName());
	private HashMap channels = new HashMap();
	private TCPDispatcher tcpDispatcher;
	private UDPDispatcher udpDispatcher;

	private class Cleaner implements Runnable {
		public void run() {
			try {
				while (true) {
					Thread.sleep(CLEANING_INTERVAL);
					ArrayList ch = new ArrayList();
					synchronized (channels) {
						for (Iterator i = channels.values().iterator();
								i.hasNext(); ) {
							ch.add(i.next());
						}
					}
					for (Iterator i = ch.iterator(); i.hasNext(); ) {
						HashMap tmpmap = (HashMap) i.next();
						ArrayList cl = new ArrayList();
						synchronized (tmpmap) {
							for (Iterator j = tmpmap.values().iterator();
									j.hasNext(); ) {
								cl.add(j.next());
							}
						}
						for (Iterator j = cl.iterator(); j.hasNext(); ) {
							ReflectorClient rc = (ReflectorClient) j.next();
							if ((System.currentTimeMillis()
									- rc.getLastAccessTime())
									> CLIENT_TIMEOUT) {
								logger.info("EXPIRE CLIENT=>"
									+ "[" + rc.getSocket()
									+ ":" + rc.getUDPAddress()
									+ ":" + rc.getUDPPort() + "]"
									+ "(" + rc.getChannel()
									+ "," + rc.getId() + ")");
								synchronized (tmpmap) {
									tmpmap.remove(new Integer(rc.getId()));
								}
								rc.close();
							}
						}
					}
				}
			} catch (Exception ex) {
				logger.severe("Cleaner thread died.");
				ex.printStackTrace();
			}
		}
	}

	private class TCPDispatcher implements Runnable {
		private ServerSocket sSocket;

		private class SocketListener extends Thread {
			private static final int BUF_SIZE = 65536;
			private Socket socket;

			SocketListener(Socket s) {
				this.socket = s;
			}

			public void run() {
				try {
					byte[] buffer = new byte[BUF_SIZE];
					DataInputStream din
						= new DataInputStream(socket.getInputStream());
					while (true) {
						int length = din.readUnsignedShort();
						din.readFully(buffer, 0, length);
						ReflectorPacket rp = new ReflectorPacket(socket,
							buffer, length);
						dispatch(rp);
					}
				} catch (Exception ex) {
					logger.info("SocketListener done. cause->" + ex);
					logger.info("  closed socket=" + socket);
				} finally {
					try {
						socket.close();
					} catch (Exception ex) {}
				}
			}
		}

		TCPDispatcher(int port) throws SocketException, IOException {
			sSocket = new ServerSocket(port);
		}

		public void run() {
			try {
				while (true) {
					Socket socket = sSocket.accept();
					new SocketListener(socket).start();
				}
			} catch (Exception ex) {
				logger.severe("Exception occurred at TCPDispatcher.");
				ex.printStackTrace();
			}
		}
	}

	private class UDPDispatcher implements Runnable {
		private DatagramSocket dSocket;

		UDPDispatcher(int port) throws SocketException {
			dSocket = new DatagramSocket(port);
		}
		
		public void run() {
			try {
				byte[] buf = new byte[BUF_SIZE];
				DatagramPacket packet = new DatagramPacket(buf, buf.length);
				while (true) {
					dSocket.receive(packet);
					ReflectorPacket rp = new ReflectorPacket(packet);
					dispatch(rp);
				}
			} catch (Exception ex) {
				logger.severe("Exception occurred at UDPDispatcher.");
				ex.printStackTrace();
			}
		}

		public DatagramSocket getDatagramSocket() {
			return dSocket;
		}
	}

	private void dispatch(ReflectorPacket rp) {
		if (logger.isLoggable(Level.FINE))
			logger.fine("RECV=" + rp);
		try {
			switch (rp.getCommand()) {
			case ENTRY:
				entry(rp);
				break;
			case SEND:
				send(rp);
				break;
			case LIST:
				list(rp);
				break;
			case ADDRESS:
				address(rp);
				break;
			case EXIT:
				exit(rp);
				break;
			default:
				break;
			}
		} catch (IOException ex) {
			logger.warning("Cannot send packet.");
			ex.printStackTrace();
		} catch (Exception ex) {
			logger.info("BAD PACKET=>" + ex);
			ex.printStackTrace();
		}
	}

	private ReflectorClient getClient(int channel, int id) {
		HashMap clients = null;
		synchronized (channels) {
			clients = (HashMap) channels.get(new Integer(channel));
		}
		if (clients == null)
			return null;
		ReflectorClient client = null;
		synchronized (clients) {
			client = (ReflectorClient) clients.get(new Integer(id));
		}
		return client;
	}

	private void entry(ReflectorPacket packet) throws IOException {
		HashMap clients = null;
		synchronized (channels) {
			Integer c = new Integer(packet.getChannel());
			clients = (HashMap) channels.get(c);
			if (clients == null) {
				clients = new HashMap();
				channels.put(c, clients);
			}
		}
		int i = 0;
		synchronized (clients) {
			for ( ; i < 0x100; i++) {
				if (!clients.containsKey(new Integer(i)))
					break;
			}
		}
		if (i == 0x100) {
			logger.warning("Channel " + packet.getChannel()
				+ " is full.");
			return;
		}
		ReflectorClient client;
		if (packet.getSocket() != null)
			client = new ReflectorClient(packet.getChannel(), i,
				udpDispatcher.getDatagramSocket(), packet.getSocket());
		else
			client = new ReflectorClient(packet.getChannel(), i,
				udpDispatcher.getDatagramSocket(), packet.getUDPAddress(),
				packet.getUDPPort());
		synchronized (clients) {
			clients.put(new Integer(i), client);
		}
		packet.setId(i);
		client.send(packet);
	}

	private byte[] getAllTargets(int channel) {
		HashMap clients = null;
		synchronized (channels) {
			clients = (HashMap) channels.get(new Integer(channel));
		}
		if (clients == null)
			return null;
		byte[] targets;
		synchronized (clients) {
			int j = 0;
			targets = new byte[clients.size()];
			for (Iterator k = clients.values().iterator();
					k.hasNext(); ) {
				ReflectorClient rc = (ReflectorClient) k.next();
				targets[j] = (byte) rc.getId();
				j++;
			}
		}
		return targets;
	}

	private void send(ReflectorPacket packet) throws IOException {
		ReflectorClient sender
			= getClient(packet.getChannel(), packet.getId());
		// valid sender?
		if (sender == null || !sender.isOwnPacket(packet)) {
			logger.info("Invalid sender.=>" + packet);
			return;
		}
		byte[] targets = packet.getTargets();
		if (targets.length == 0) { // multicast to all members.
			targets = getAllTargets(packet.getChannel());
			if (targets == null)
				return;
		}
		for (int j = 0; j < targets.length; j++) {
			int target = (targets[j] < 0)
				? targets[j] + 0x100 : targets[j];
			ReflectorClient rc = getClient(packet.getChannel(), target);
			if (rc == null)
				continue;
			rc.send(packet);
		}
	}

	private void list(ReflectorPacket packet) throws IOException {
		ReflectorClient sender
			= getClient(packet.getChannel(), packet.getId());
		// valid sender?
		if (sender == null || !sender.isOwnPacket(packet)) {
			logger.info("Invalid sender.=>" + packet);
			return;
		}
		byte[] targets = getAllTargets(packet.getChannel());
		if (targets == null)
			return;
		packet.setTargets(targets);
		sender.send(packet);
	}

	private void address(ReflectorPacket packet) throws IOException {
		ReflectorClient sender
			= getClient(packet.getChannel(), packet.getId());
		// valid sender?
		if (sender == null || !sender.isOwnPacket(packet)) {
			logger.info("Invalid sender.=>" + packet);
			return;
		}
		byte[] targets = packet.getTargets();
		if (targets == null || targets.length == 0)
			return;
		ByteArrayOutputStream bout = new ByteArrayOutputStream();
		DataOutputStream dout = new DataOutputStream(bout);
		for (int i = 0; i < targets.length; i++) {
			int id = (targets[i] < 0) ? targets[i] + 0x100 : targets[i];
			ReflectorClient tmprc = getClient(packet.getChannel(), id);
			if (tmprc == null || tmprc.getUDPAddress() == null)
				continue;
			dout.writeByte(id);
			dout.write(tmprc.getUDPAddress().getAddress());
			dout.writeShort(tmprc.getUDPPort());
		}
		dout.close();
		packet.setData(bout.toByteArray());
		sender.send(packet);
	}

	private void exit(ReflectorPacket packet) throws IOException {
		ReflectorClient sender
			= getClient(packet.getChannel(), packet.getId());
		// valid sender?
		if (sender == null || !sender.isOwnPacket(packet)) {
			logger.info("Invalid sender.=>" + packet);
			return;
		}
		HashMap clients = null;
		synchronized (channels) {
			clients = (HashMap) channels.get(new Integer(sender.getChannel()));
		}
		if (clients == null)
			return;
		synchronized (clients) {
			clients.remove(new Integer(sender.getId()));
		}
		sender.close();
	}

	public Reflector(int port) throws SocketException, IOException {
		udpDispatcher = new UDPDispatcher(port);
		Thread d1 = new Thread(udpDispatcher);
		d1.start();
		tcpDispatcher = new TCPDispatcher(port);
		Thread d2 = new Thread(tcpDispatcher);
		d2.start();
		Cleaner c = new Cleaner();
		Thread cleaner = new Thread(c);
		cleaner.start();
	}

}
